"""version-related helpers"""

from distutils.version import StrictVersion, Version
from typing import Union, Optional


class StrictVersionHelper(StrictVersion):
    """internal helper 'coz tuple juggling is no fun!"""

    def __init__(self, vstring: Union[Version, str] = None) -> None:
        super().__init__(str(vstring) if vstring else None)

    @property
    def major(self) -> int:
        return self.version[0]

    @major.setter
    def major(self, value: int) -> None:
        self.version = (value, self.minor, self.patch)

    @property
    def minor(self) -> int:
        return self.version[1]

    @minor.setter
    def minor(self, value: int) -> None:
        self.version = (self.major, value, self.patch)

    @property
    def patch(self) -> int:
        return self.version[2]

    @patch.setter
    def patch(self, value: int) -> None:
        self.version = (self.major, self.minor, value)

    @property
    def prerelease_stage(self) -> Optional[str]:
        return self.prerelease[0] if self.prerelease else None

    @prerelease_stage.setter
    def prerelease_stage(self, value: Optional[str]) -> None:
        self.prerelease = None if value is None else (value, self.prerelease_num)

    @property
    def prerelease_num(self) -> Optional[int]:
        return self.prerelease[1] if self.prerelease else None

    @prerelease_num.setter
    def prerelease_num(self, value: Optional[int]) -> None:
        self.prerelease = None if value is None else (self.prerelease_stage, value)

    def bump_next_release(self) -> None:
        """Bumps the current version to the next release."""
        if not self.prerelease:
            # 1.4.4 bumps to 1.4.5
            self.patch += 1
        else:
            ...  # nothing do to: 1.4.5a3 would become 1.4.5

        self.prerelease = None

    def bump_next_prerelease(self, *, patch: bool = True) -> None:
        """Bumps the current version to the next prerelease."""
        if self.prerelease:
            # 1.4.5a4 bumps to 1.4.5a5
            self.prerelease_num += 1
        else:
            # 1.4.4 bumps to 1.4.5a1
            self.patch += int(patch)  # either 1 or 0
            self.prerelease = ("a", 1)

    def __copy__(self) -> "StrictVersionHelper":
        return StrictVersionHelper(self)



from collections import defaultdict
from distutils.version import LooseVersion
from typing import List, Dict

import click
from coveo_styles.styles import echo, ExitWithFailure, install_pretty_exception_hook

from .exceptions import VersionExists
from .pypi import (
    obtain_versions_from_pypi,
    compute_next_version,
    obtain_latest_release_from_pypi,
    PYPI_CLI_INDEX,
)
from .versions import StrictVersionHelper


click_argument_package = click.argument("package")
click_option_index = click.option(
    "--index",
    default=str(PYPI_CLI_INDEX),
    help='The pypi index host, in the form "https://pypi.org"',
)


@click.group()
def pypi() -> None:
    install_pretty_exception_hook()


@pypi.command()
@click_argument_package
@click.argument("version")
@click_option_index
def raise_if_exists(package: str, version: str, index: str = str(PYPI_CLI_INDEX)) -> None:
    """Raise if the version already exists."""
    if version in obtain_versions_from_pypi(package, index=index):
        raise ExitWithFailure(
            suggestions='Bump the version using "poetry version major|patch|etc" and retry.'
        ) from VersionExists(f"{package}=={version}")


@pypi.command()
@click_argument_package
@click_option_index
def versions(package: str, index: str = str(PYPI_CLI_INDEX)) -> None:
    """Prints the (unsorted) versions of a package."""
    groups: Dict[str, List[LooseVersion]] = defaultdict(list)
    for loose_version in obtain_versions_from_pypi(
        package, version_class=LooseVersion, oldest_first=True, index=index
    ):
        version = str(loose_version)
        if "." not in version:
            groups["uncanny"].append(loose_version)
        else:
            groups[version.split(".")[0]].append(loose_version)

    for group, package_versions in groups.items():
        echo.normal(f"\nVersion {group}:")
        for loose_version in package_versions:
            echo.noise(loose_version, item=True)


@pypi.command()
@click_argument_package
@click.option("--prerelease", is_flag=True, default=False)
@click.option("--minimum-version", default="0.0.1")
@click_option_index
def next_version(
    package: str,
    prerelease: bool = False,
    minimum_version: str = "0.0.1",
    index: str = str(PYPI_CLI_INDEX),
) -> None:
    """Returns the version number for this release."""
    echo.passthrough(
        compute_next_version(
            package, prerelease=prerelease, minimum_version=minimum_version, index=index
        )
    )


@pypi.command()
@click_argument_package
@click_option_index
def current_version(package: str, index: str = str(PYPI_CLI_INDEX)) -> None:
    """Returns the most recent official release version."""
    echo.passthrough(
        obtain_latest_release_from_pypi(package, index=index) or StrictVersionHelper("0.0.0")
    )



from distutils.version import Version, StrictVersion
from typing import List, Type, TypeVar, Optional

from coveo_settings.settings import StringSetting
import requests

from .exceptions import VersionException
from .versions import StrictVersionHelper


PYPI_CLI_INDEX = StringSetting("pypi.cli.index", fallback="https://pypi.org")

T = TypeVar("T", bound=Version)


class VersionExists(Exception):
    ...


def obtain_versions_from_pypi(
    package_name: str,
    index: str = str(PYPI_CLI_INDEX),
    *,
    version_class: Type[T] = StrictVersion,  # type: ignore
    oldest_first: bool = False,
) -> List[T]:
    """
    Requests all versions of a package from a pypi server.

    oldest_first: sort order
    version_class: some functionality depends on StrictVersion. LooseVersion may be used to obtain
      packages that don't follow distutils' best practices.
    """
    response = requests.get(f"{index}/pypi/{package_name}/json")
    if response.status_code == 404:
        return []  # no hits; that might be ok.
    response.raise_for_status()
    data = response.json()

    valid_versions: List[T] = []
    for version in data["releases"].keys():
        try:
            valid_versions.append(version_class(version))
        except ValueError:
            pass  # invalid under this scheme

    # no need for a generator, sorting requires all results anyway.
    try:
        return sorted(valid_versions, reverse=not oldest_first)
    except TypeError:  # happens when versions are not standard (like dev1); use str sort :shrug:
        return sorted(valid_versions, reverse=not oldest_first, key=str)


def obtain_latest_release_from_pypi(
    package: str, index: str = str(PYPI_CLI_INDEX)
) -> Optional[StrictVersion]:
    """Obtains the latest non-prerelease version from pypi."""
    official_releases = filter(
        lambda version: not version.prerelease,
        obtain_versions_from_pypi(package, version_class=StrictVersionHelper, index=index),
    )
    return next(official_releases, None)


def compute_next_version(
    package: str,
    *,
    prerelease: bool,
    minimum_version: str = "0.0.1",
    index: str = str(PYPI_CLI_INDEX),
) -> StrictVersionHelper:
    """
    Given a package, compute the next version based on what's in pypi. e.g.:
        - If 1.0.0 is the latest release version, it becomes 1.0.1 or 1.0.1a1
        - If 1.0.1a1 is the latest prerelease, it becomes 1.0.1 or 1.0.1a2

    If a minimum version is specified and is higher than the latest release, it is considered the next version to
    release. e.g.: If 1.0.0 is the minimum version and the latest is lower:
            - 1.0.0 is the next version
            - 1.0.0a1 is the next prerelease
    """
    lbound_version = StrictVersionHelper(minimum_version)
    if lbound_version.prerelease:
        raise VersionException(f"Minimum version {minimum_version} cannot be a pre-release.")
    if lbound_version < StrictVersionHelper("0.0.1"):
        raise VersionException(f"Minimum version {minimum_version} must be 0.0.1 or higher.")

    # don't bump the patch number of the minimum version, but process the prerelease bump if applicable.
    if prerelease:
        lbound_version.bump_next_prerelease(patch=False)

    latest_release = next(
        map(StrictVersionHelper, obtain_versions_from_pypi(package, index=index)), None
    )
    if latest_release is None:
        return lbound_version

    if prerelease:
        latest_release.bump_next_prerelease()
    else:
        latest_release.bump_next_release()

    return max(lbound_version, latest_release)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['coveo_pypi_cli']

package_data = \
{'': ['*']}

install_requires = \
['click',
 'coveo-settings>=2.0.0,<3.0.0',
 'coveo-styles>=2.0.0,<3.0.0',
 'coveo-systools>=2.0.0,<3.0.0',
 'requests',
 'typing-extensions']

entry_points = \
{'console_scripts': ['pypi = coveo_pypi_cli.cli:pypi']}

setup_kwargs = {
    'name': 'coveo-pypi-cli',
    'version': '2.0',
    'description': 'Query and compute pypi versions from command line.',
    'long_description': '# coveo-pypi-cli\n\nA very simple pypi cli that can be used to obtain the latest version of a package, or calculate the next one.\n\nServes our automatic pypi push github action.\n\n\n## `pypi current-version`\n\nDisplay the current version of a package from `pypi.org`\n\n\n## `pypi next-version`\n\nCompute the next version of a package.\n\n- Can be given a minimum version\n  - e.g.: pypi is `0.0.3` and mininum set to `0.1`: next version will be `0.1`\n- Supports computing pre-release versions\n\n# private index support\n\nYou can target a private pypi server through a switch or an environment variable.\n\n## Using the `--index` switch\n\n```shell\n$ pypi current-version secret-package --index https://my.pypi.server.org\n1.0.0\n\n$ pypi current-version secret-package --index https://my.pypi.server.org:51800/urlprefix\n1.0.0\n```\n\n## Using the environment variable:\n\n```shell\n$ PYPI_CLI_INDEX="https://my.pypi.server.org" pypi current-version secret-package\n```\n\nNote: Unlike `pip --index-url`, **you must omit** the `/simple` url prefix.\nThe API used by `coveo-pypi-cli` is served by the `/pypi` endpoint _and should not be specified either!_\n\n\n# pypi-cli in action\n\nThe best example comes from the [github action](./.github/workflows/actions/publish-to-pypi), which computes the next version based on the current release and what\'s in the `pyproject.toml`.\n\nHere\'s what you can expect from the tool:\n\n```shell\n$ pypi current-version coveo-functools\n0.2.1\n\n$ pypi next-version coveo-functools\n0.2.2\n\n$ pypi next-version coveo-functools --prerelease\n0.2.2a1\n\n$ pypi next-version coveo-functools --minimum-version 0.2\n0.2.2\n\n$ pypi next-version coveo-functools --minimum-version 0.3\n0.3\n\n$ pypi next-version coveo-functools --minimum-version 0.3 --prerelease\n0.3a1\n\n\n# Here\'s an example of how we use it in the github action\n\n$ poetry version\ncoveo-pypi-cli 0.1.0\n$ minimum_version=$(poetry version | cut --fields 2 --delimiter \' \' )\n0.1.0\n\n# when left unattended, the next-version increments the patch number\n$ pypi next-version coveo-pypi-cli --minimum-version $minimum_version\n0.2.2\n\n# in order to change the minor or major, because the script uses `poetry version` to obtain the minimum version, \n# just set it in `pyproject.toml` manually or by calling `poetry version <new-version>` (and commit!)\n$ poetry version 0.3\nBumping version from 0.1.0 to 0.3\n$ minimum_version=$(poetry version | cut --fields 2 --delimiter \' \' )\n0.3\n$ pypi next-version coveo-pypi-cli --minimum-version $minimum_version\n0.3\n\n# IMPORTANT: the publish step MUST set the computed version for poetry before publishing!\n$ poetry version $minimum_version\n0.3\n$ poetry publish\n...\n\n# after publishing the above, repeating the steps would yield:\n$ pypi next-version coveo-pypi-cli --minimum-version $minimum_version\n0.3.1\n\n# for completeness, you can also publish pre-releases:\n$ pypi next-version coveo-pypi-cli --minimum-version $minimum_version --prerelease\n0.3.1a1\n\n \n```\n',
    'author': 'Jonathan Piché',
    'author_email': 'tools@coveo.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/coveooss/coveo-python-oss',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
